#include <cassert>
#include <chrono>
#include <iostream>
#include <mutex>
#include <system_error>  
#include <thread>

#include "Illustration_of_thread_local.h"

thread_local const std::string Illustration_of_thread_local::_Francois = "François";
thread_local const std::string Illustration_of_thread_local::_Hollande = "Hollande";
thread_local const std::string Illustration_of_thread_local::_Nicolas = "Nicolas";
thread_local const std::string Illustration_of_thread_local::_Sarkozy = "Sarkozy";

std::string Illustration_of_thread_local::_Patronym = "?";
bool Illustration_of_thread_local::_Checkable = false;
bool Illustration_of_thread_local::_Stop = false;

int Illustration_of_thread_local::Bads = 0;
int Illustration_of_thread_local::Goods = 0;

std::mutex Illustration_of_thread_local::_Mutex;

void Illustration_of_thread_local::_Check() {
	while (!_Stop) {
		if (_Checkable) {
			// 'thread_local' for '_Francois', '_Hollande', '_Nicolas' and '_Sarkozy':
			if (_Patronym.compare(_Francois + _Sarkozy) == 0 || _Patronym.compare(_Nicolas + _Hollande) == 0)
				Bads++;
			else
				Goods++;
			_Checkable = false;
		}
	}
}

void Illustration_of_thread_local::_Set(std::string const& given_name, std::string const& surname) {
	/* Processing is *deliberately* slowed down through two statements instead of '_Patronym = given_name + surname;' */
	// const std::lock_guard<std::mutex> critical_section(_Mutex); // Lightweight...
	std::unique_lock<std::mutex> critical_section(_Mutex); // Heavyweight...
	// critical_section.try_lock(); // Execution error since constructor does it just before...
	_Patronym = given_name; // Random segmentation fault if no critical section set up...
	critical_section.unlock();
	critical_section.lock();
	_Patronym += surname; // Random segmentation fault if no critical section set up...
	critical_section.unlock();
} // Mutex is released through destruction of 'critical_section' object...

void Illustration_of_thread_local::_Set_Francois_Hollande() {
	while (!_Stop) {
		_Set(_Francois, _Hollande); // 'thread_local' for '_Francois', '_Hollande'...
		_Checkable = true;
	}
}

void Illustration_of_thread_local::_Set_Nicolas_Sarkozy() {
	while (!_Stop) {
		_Set(_Nicolas, _Sarkozy); // 'thread_local' for '_Nicolas' and '_Sarkozy'...
		_Checkable = true;
	}
}

void Illustration_of_thread_local::Go() {
	std::cout << "Concurrent supported threads: " << std::thread::hardware_concurrency() << '\n';

	std::thread Check(&Illustration_of_thread_local::_Check);
	std::thread Set_Francois_Hollande(&Illustration_of_thread_local::_Set_Francois_Hollande);
	std::thread Set_Nicolas_Sarkozy(&Illustration_of_thread_local::_Set_Nicolas_Sarkozy);

	assert(Check.joinable()); // Still joinable if execution code already finished...
	assert(Set_Francois_Hollande.joinable());
	assert(Set_Nicolas_Sarkozy.joinable());

	std::this_thread::sleep_for(std::chrono::milliseconds(10));
	_Stop = true;

	Check.join();
	Set_Francois_Hollande.join();
	Set_Nicolas_Sarkozy.join();
}

int main(int argc, char** argv) {

	enum class Direction : char {
		East = 'E', South = 'S', West = 'W', North = 'N'
	};
	assert(sizeof(Direction) == 1);
	// Using related 'std::underlying_type' trait:
	static_assert(std::is_same<std::underlying_type<Direction>::type, char>::value, "'Direction' has not 'char' as underlying type, why?");

	Illustration_of_thread_local::Go();
	std::cout << "Bads compositions: " << Illustration_of_thread_local::Bads << '\n';
	std::cout << "Goods compositions: " << Illustration_of_thread_local::Goods << '\n';
	std::cout << "Total: " << Illustration_of_thread_local::Bads + Illustration_of_thread_local::Goods << '\n';

	bool stop = false;
	// Capture 'stop' by reference for guaranteed stopping:
	std::thread my_thread([&stop]() -> void { // Thread creation from lambda expression...
		while (!stop) {
		}
		std::cout << "Just ends..." << '\n';
		});
	assert(my_thread.joinable());
	my_thread.detach();
	assert(!my_thread.joinable());
	std::this_thread::sleep_for(std::chrono::seconds(1));
	stop = true;
	try {
		my_thread.join();
	}
	catch (std::system_error& se) { // Inheritance graph: 'std::exception' < 'std::runtime_error' < 'std::system_error' (C++11)
		std::cerr << "'my_thread.join()' failed because 'my_thread.joinable()' is " << std::boolalpha << my_thread.joinable() << ", message: " << se.code().message()  << '\n';
	}

	return 0;
}

